package de.hsbo.fbv.bmg.geometry.simple;

import java.awt.Shape;
import java.awt.geom.Path2D;
import java.util.ArrayList;

import com.vividsolutions.jts.geom.Coordinate;

import de.hsbo.fbv.ogc.geometry.simple.Geometry;
import de.hsbo.fbv.ogc.geometry.simple.LineString;
import de.hsbo.fbv.ogc.geometry.simple.Point;

/**
 * Implementierung des Interface LineString auf Basis der Klassenbibliothek Java
 * Topology Suite (JTS) 1.10 der Fa. VividSolutions.<br>
 * (C) siehe Lizenzen von Geotools <br>
 * 
 * @author A.Zimmermann, HS Bochum, 2011
 * @version 0.5
 */
public class GLineString extends GGeometry implements LineString {

	/** Attribut jtsGeomColl : Delegat-Objekt zum Zugriff auf die JTS */
	protected com.vividsolutions.jts.geom.LineString jtsLineString = null;

	/**
	 * Konstruktor
	 * @param points
	 * @throws Exception
	 */
	public GLineString(GPoint[] points) throws Exception {
		int npts = points.length;
		CoordinateM[] coords = new CoordinateM[npts];
		for (int i = 0; i < npts; i++) {
			coords[i] = new CoordinateM(points[i].getJtsPoint().getCoordinate());
		}
		this.jtsLineString = GeoModel.FACTORY.createLineString(coords);
		if (!this.jtsLineString.isValid())
			throw new Exception("Object not valid!");
		this.shape = getShape(coords);
	}

	/**
	 * Konstruktor
	 * @param coords
	 * @throws Exception
	 */
	public GLineString(CoordinateM[] coords) throws Exception {
		this.jtsLineString = GeoModel.FACTORY.createLineString(coords);
		if (!this.jtsLineString.isValid())
			throw new Exception("Object not valid!");
		this.shape = getShape(coords);
	}

	// Implementierte Methoden von LineString

	@Override
	public Point endPoint() {
		return GeoModel.createPoint(getJtsLineString().getEndPoint());
	}

	@Override
	public boolean isClosed() {
		return getJtsLineString().isClosed();
	}

	@Override
	public boolean isRing() {
		return getJtsLineString().isRing();
	}

	@Override
	public double length() {
		return getJtsLineString().getLength();
	}

	@Override
	public Point startPoint() {
		return GeoModel.createPoint(getJtsLineString().getStartPoint());
	}

	@Override
	public int numPoints() {
		return getJtsLineString().getNumPoints();
	}

	@Override
	public Point pointN(int n) {
		return GeoModel.createPoint(getJtsLineString().getPointN(n));
	}

	@Override
	public com.vividsolutions.jts.geom.LineString toJtsGeom() {
		return this.jtsLineString;
	}

	@Override
	public int coordinateDimension() {
		Coordinate[] coords = getJtsLineString().getCoordinates();
		int cdim = 3;
		for (int i = 0; i < coords.length; i++) {
			cdim = Math.min(cdim,
					((CoordinateM) coords[i]).getCoordinateDimension());
		}
		return cdim;
	}

	@Override
	public boolean isMeasured() {
		Coordinate[] coords = getJtsLineString().getCoordinates();
		boolean res = true;
		for (int i = 0; i < coords.length; i++) {
			res = res && ((CoordinateM) coords[i]).isMeasured();
		}
		return res;
	}

	/** Noch nicht implementiert! */
	@Override
	public Geometry locateAlong(double arg0) {
		try {
			throw new Exception("not yet implemented");
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/** Noch nicht implementiert! */
	@Override
	public Geometry locateBetween(double arg0, double arg1) {
		try {
			throw new Exception("not yet implemented");
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Interne Methode:<br>
	 * Bildet aus der vorliegenden Geometrie ein Shape-Objekt (Typ Path2D).
	 * @param coords
	 * @return
	 */
	private Shape getShape(Coordinate[] coords) {
		Path2D path = new Path2D.Double();
		path.moveTo(coords[0].x, coords[0].y);
		for (int i = 1; i < coords.length; i++)
			path.lineTo(coords[i].x, coords[i].y);
		return path;
	}

	/**
	 * Gibt das Delegat-Objekt zurck.<br>
	 * Achtung: Das Objekt kann unbeabsichtigt verndert werden!
	 * @return
	 */
	public com.vividsolutions.jts.geom.LineString getJtsLineString() {
		return jtsLineString;
	}

	/**
	 * Konvertiert dieses Objekt in ein Koordinaten-Feld (Typ CoordinateM)
	 * @return
	 */
	public CoordinateM[] toCoordinateArray() {
		return CoordinateM.toCoordinateArray(jtsLineString.getCoordinates());
	}

	/**
	 * Konvertiert dieses Objekt in eine ArrayList von Typ CoordinateM.
	 * @return
	 */
	public ArrayList<CoordinateM> toCoordinateList() {
		return CoordinateM.toCoordinateList(jtsLineString.getCoordinates());
	}

	/**
	 * Konvertiert dieses Objekt in ein Feld seiner Streckensegmente (Typ GLine).
	 * @return
	 */
	public GLine[] toLines() {
		try {
			ArrayList<GLine> lines = new ArrayList<GLine>();
			Coordinate[] coords = jtsLineString.getCoordinates();
			int n = coords.length;
			for (int i = 0; i < n - 1; i++) {
				lines.add(new GLine((CoordinateM) coords[i],
						(CoordinateM) coords[i + 1]));
			}
			/*
			if (this.isClosed())
				lines.add(new GLine((CoordinateM) coords[n - 1],
						(CoordinateM) coords[0]));
						*/
			GLine[] lineA = new GLine[lines.size()];
			return lines.toArray(lineA);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Konvertiert dieses Objekt in ein GLinearRing-Objekt.<br>
	 * Der LineString darf keine Selbst-berschneidungen enthalten. Ein fehlender 
	 * Abschlusspunkt wird ggf. ergnzt, d.h der Linienzug wird geschlossen.
	 * @return
	 */
	public GLinearRing toLinearRing() {
		try {
			if (!isSimple())
				throw new Exception(
						"LineString not simple - cannot create ring!");
			if (isClosed()) {
				return new GLinearRing(this.toCoordinateArray());
			} else {
				CoordinateM[] src = toCoordinateArray();
				CoordinateM[] dst = new CoordinateM[src.length + 1];
				System.arraycopy(src, 0, dst, 0, src.length);
				dst[dst.length - 1] = src[0];
				return new GLinearRing(dst);
			}
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}
	
	/**
	 * Gibt die Punkte eines LineString-Objektes aus.<br>
	 * Die Methode verwendet intern numPoints() und PointN().
	 * @return
	 */
	public GPoint[] getPoints() {
		GPoint[] pts = new GPoint[numPoints()];
		for(int i=0; i<numPoints(); i++)
			pts[i] = (GPoint)(this.pointN(i));
		return pts;
	}
	
	

}
