package de.hsbo.fbv.bmg.geometry.simple;

import java.awt.Shape;
import java.awt.geom.Path2D;
import java.util.Arrays;
import java.util.List;

import de.hsbo.fbv.ogc.geometry.simple.Geometry;
import de.hsbo.fbv.ogc.geometry.simple.LinearRing;
import de.hsbo.fbv.ogc.geometry.simple.Point;
import de.hsbo.fbv.ogc.geometry.simple.Polygon;

/**
 * Implementierung des Interface Polygon auf Basis der Klassenbibliothek Java
 * Topology Suite (JTS) 1.10 der Fa. VividSolutions.<br>
 * (C) siehe Lizenzen von Geotools <br>
 * 
 * @author A.Zimmermann, HS Bochum, 2011
 * @version 0.5
 */
public class GPolygon extends GGeometry implements Polygon {

	/** Attribut jtsGeomColl : Delegat-Objekt zum Zugriff auf die JTS */
	protected com.vividsolutions.jts.geom.Polygon jtsPolygon = null;

	/**
	 * Konstruktor
	 * @param shell
	 * @param holes
	 * @throws Exception
	 */
	public GPolygon(GLinearRing shell, GLinearRing[] holes) throws Exception {
		com.vividsolutions.jts.geom.LinearRing jshell = shell.toJtsLinearRing();
		com.vividsolutions.jts.geom.LinearRing[] jholes = null;
		if (holes != null) {
			jholes = new com.vividsolutions.jts.geom.LinearRing[holes.length];
			for (int i = 0; i < holes.length; i++) {
				jholes[i] = holes[i].toJtsLinearRing();
			}
		}
		jtsPolygon = GeoModel.FACTORY.createPolygon(jshell, jholes);
		if (!jtsPolygon.isValid())
			throw new Exception("Object not valid!");
		this.shape = getShape(shell, holes);
	}

	// Implementierte Methoden

	@Override
	public double area() {
		return getJtsPolygon().getArea();
	}

	@Override
	public Point centroid() {
		return GeoModel.createPoint(getJtsPolygon().getCentroid());
	}

	@Override
	public Point pointOnSurface() {
		return GeoModel.createPoint(getJtsPolygon().getInteriorPoint());
	}

	@Override
	public LinearRing exteriorRing() {
		return GeoModel.createLinearRing(getJtsPolygon().getExteriorRing());
	}

	@Override
	public LinearRing interiorRingN(int n) {
		return GeoModel.createLinearRing(getJtsPolygon().getInteriorRingN(n));
	}

	@Override
	public int numInteriorRing() {
		return getJtsPolygon().getNumInteriorRing();
	}

	@Override
	public com.vividsolutions.jts.geom.Polygon toJtsGeom() {
		return this.jtsPolygon;
	}

	@Override
	public int coordinateDimension() {
		int cdim = this.exteriorRing().coordinateDimension();
		for (int i = 0; i < this.numInteriorRing(); i++) {
			cdim = Math.min(cdim, this.interiorRingN(i).coordinateDimension());
		}
		return cdim;
	}

	@Override
	public boolean isMeasured() {
		boolean res = this.exteriorRing().isMeasured();
		for (int i = 0; i < this.numInteriorRing(); i++) {
			res = res && this.interiorRingN(i).isMeasured();
		}
		return res;
	}

	/** Noch nicht implementiert! */
	@Override
	public Geometry locateAlong(double arg0) {
		try {
			throw new Exception("not yet implemented");
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/** Noch nicht implementiert! */
	@Override
	public Geometry locateBetween(double arg0, double arg1) {
		try {
			throw new Exception("not yet implemented");
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	// zustzliche Methoden

	/**
	 * Interne Methode:<br>
	 * Bildet aus der vorliegenden Geometrie ein Shape-Objekt (Typ Path2D).
	 * @param shell
	 * @param holes
	 * @return
	 */
	private Shape getShape(GLinearRing shell, GLinearRing[] holes) {
		Path2D path = new Path2D.Double();
		path.append(shell, false);
		if (holes != null) {
			for (int i = 0; i < holes.length; i++) {
				path.append(holes[i], false);
			}
		}
		return path;
	}
	
	/**
	 * Gibt das Delegat-Objekt zurck.<br>
	 * Achtung: Das Objekt kann unbeabsichtigt verndert werden!
	 * @return
	 */
	public com.vividsolutions.jts.geom.Polygon getJtsPolygon() {
		return jtsPolygon;
	}

	/**
	 * Konvertiert dieses Objekt in ein Feld seiner Streckensegmente (Typ GLine).
	 * @return
	 */
	public GLine[] toLines() {
		GLinearRing ring = (GLinearRing) this.exteriorRing();
		List<GLine> result = Arrays.asList(ring.toLines());
		for (int i = 0; i < numInteriorRing(); i++) {
			GLinearRing ringi = (GLinearRing) this.interiorRingN(i);
			result.addAll(Arrays.asList(ringi.toLines()));
		}
		GLine[] lines = new GLine[result.size()];
		return result.toArray(lines);
	}

	@Override
	public String toString() {
		return this.asText();
	}

}
