package de.hsbo.fbv.bmg.geometry.simple;

import java.awt.Shape;
import java.awt.geom.Ellipse2D;
import java.util.ArrayList;
import java.util.List;

import com.vividsolutions.jts.geom.Coordinate;

import de.hsbo.fbv.ogc.geometry.simple.Triangle;

/**
 * Implementierung des Interface Triangle auf Basis der Klassenbibliothek Java
 * Topology Suite (JTS) 1.10 der Fa. VividSolutions.<br>
 * (C) siehe Lizenzen von Geotools <br>
 * Die Klasse greift auf verschiedene statische Methoden der Klasse Triangle von
 * JTS zurck.<br>
 * 
 * @author A.Zimmermann, HS Bochum, 2011
 * @version 0.5
 */
public class GTriangle extends GPolygon implements Triangle {

	// Parameter
	public static final int A = 0;
	public static final int B = 1;
	public static final int C = 2;
	public static final int AB = 2;
	public static final int BC = 0;
	public static final int CA = 1;

	// Attribute
	XCircle circle = null; // Umkreis
	GTriangle[] nextTris = new GTriangle[3];

	/**
	 * Konstruktor mit Angabe der Dreieckspunkte A,B,C
	 * 
	 * @param a
	 * @param b
	 * @param c
	 * @throws Exception
	 */
	public GTriangle(CoordinateM a, CoordinateM b, CoordinateM c)
			throws Exception {
		super(new GLinearRing(new CoordinateM[] { a, b, c, a }), null);
		Coordinate center = com.vividsolutions.jts.geom.Triangle.circumcentre(
				a, b, c);
		double radius = center.distance(a);
		this.circle = new XCircle(center, radius);
	}

	/**
	 * Gibt den Punkt (CoordinateM) mit dem Index i zurck.<br>
	 * Der Index muss zwischen 0 und 2 liegen, ansonsten wird null
	 * zurckgegeben.
	 * 
	 * @param i
	 * @return
	 */
	public CoordinateM getPoint(int i) {
		if (i >= 0 && i < 3)
			return ((GPoint) (this.exteriorRing().pointN(i))).getCoordinate();
		else
			return null;
	}

	/**
	 * Gibt den Index des Punktes p (CoordinateM) zurck.<br>
	 * Ist der Punkt nicht bekannt, wird der Index -1 ausgegeben.
	 * 
	 * @param p
	 * @return
	 */
	public int indexOf(CoordinateM p) {
		for (int i = 0; i < 3; i++) {
			if (p.equals2D(this.getPoint(i))) {
				return i;
			}
		}
		return -1;
	}

	/**
	 * Gibt den Index der Seite mit den Endpunkten p0 und p1 (CoordinateM)
	 * zurck.<br>
	 * Der Seitenindex entspricht immer dem Index des Punktes, der der Seite
	 * gegenberliegt. Ist einer der Punkte nicht bekannt, wird der Index -1
	 * ausgegeben.
	 * 
	 * @param p0
	 * @param p1
	 * @return
	 */
	public int indexOf(CoordinateM p0, CoordinateM p1) {
		int i0 = this.indexOf(p0);
		int i1 = this.indexOf(p1);
		if (i0 == -1 || i1 == -1)
			return -1;
		else
			return 3 - i0 - i1;
	}

	/**
	 * Gibt den Punkte (CoordinateM) zurck, der der angegebenen Seite
	 * gegenberliegt.<br>
	 * 
	 * @param p0
	 * @param p1
	 * @return
	 */
	public CoordinateM getOppositePoint(CoordinateM p0, CoordinateM p1) {
		return getPoint(indexOf(p0, p1));
	}

	/**
	 * Stellt fest, ob sich ein Punkt p (CoordinateM) innerhalb des Umkreises
	 * befindet.<br>
	 * Liegt p auf oder auerhalb des Kreises, wird false zurckgegeben.
	 * 
	 * @param p
	 * @return
	 */
	public boolean isInCircle(CoordinateM p) {
		return circle.isInCircle(p);
	}

	/**
	 * Gibt die gemeinsamen Punkte (CoordinateM) der beiden Dreiecke this und
	 * triangle zurck.<br>
	 * Die Ergebnisliste kann null bis drei gemeinsame Punkte enthalten.
	 * 
	 * @param triangle
	 * @return
	 */
	public List<CoordinateM> getCommonPoints(GTriangle triangle) {
		List<CoordinateM> list = new ArrayList<CoordinateM>();
		if (triangle == null)
			return null;
		for (int i = 0; i < 3; i++) {
			CoordinateM p = this.getPoint(i);
			for (int j = 0; j < 3; j++) {
				if (triangle.indexOf(p) != -1) {
					list.add(p);
					break;
				}
			}
		}
		return list;
	}

	/**
	 * Speichert einen Verweis auf das Nachbardreieck.<br>
	 * Beide Dreiecke mssen eine gemeinsame Kante haben. Die Aktion wird
	 * abgebrochen und mit false quittiert, falls die Bedingung nicht erfllt
	 * ist.
	 * 
	 * @param triangle
	 * @return
	 */
	public boolean setNextTriangle(GTriangle triangle) {
		List<CoordinateM> list = this.getCommonPoints(triangle);
		if (list == null || list.size() != 2) {
			return false;
		} else {
			int iside = indexOf(list.get(0), list.get(1));
			this.nextTris[iside] = triangle;
			return true;
		}
	}

	/**
	 * Speichert alle drei Verweise der Nachbardreiecke.<br>
	 * 
	 * @param t1
	 * @param t2
	 * @param t3
	 * @return
	 */
	public boolean setNextTriangles(GTriangle t1, GTriangle t2, GTriangle t3) {
		return setNextTriangle(t1) && setNextTriangle(t2)
				&& setNextTriangle(t3);
	}

	/**
	 * Gibt das Nachbardreieck der Seite mit dem Index i zurck.
	 * 
	 * @param i
	 * @return
	 */
	public GTriangle getNextTriangle(int i) {
		return this.nextTris[i];
	}

	/**
	 * Gibt die benachbarten Dreiecke zurck
	 * 
	 * @return
	 */
	public GTriangle[] getNextTriangles() {
		return this.nextTris;
	}

	/**
	 * Gibt eine Darstellungsgraphik des umschreibenden Kreises zurck.
	 * 
	 * @return Shape
	 */
	public Shape getCircleShape() {
		return circle.getShape();
	}

} // Ende Klasse GTriangle

/**
 * Hilfsklasse XCircle
 * 
 * @author zimmermann
 * 
 */
class XCircle {

	// Attribute
	Coordinate center = null;
	double radius = 0.0;

	// Konstruktor
	XCircle(Coordinate center, double radius) {
		this.center = center;
		this.radius = radius;
	}

	// Methoden
	boolean isInCircle(CoordinateM p) {
		double d = p.distance(center);
		if ((radius - d) > GeoModel.getEPS())
			return true;
		else
			return false;
	}

	boolean isOnCircle(CoordinateM p) {
		double d = p.distance(center);
		if (Math.abs(d - radius) < GeoModel.getEPS())
			return true;
		else
			return false;
	}

	boolean isOutOfCircle(CoordinateM p) {
		double d = p.distance(center);
		if (Math.abs(d - radius) > GeoModel.getEPS())
			return true;
		else
			return false;
	}

	Shape getShape() {
		double xmin = center.x - radius;
		double ymin = center.y - radius;
		double d = 2 * radius;
		return new Ellipse2D.Double(xmin, ymin, d, d);
	}

} // Ende Klasse XCircle