package de.hsbo.fbv.bmg.geometry.simple;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.GeometryFactory;
import de.hsbo.fbv.ogc.geometry.simple.Geometry;
import de.hsbo.fbv.ogc.geometry.simple.GeometryCollection;
import de.hsbo.fbv.ogc.geometry.simple.Line;
import de.hsbo.fbv.ogc.geometry.simple.LineString;
import de.hsbo.fbv.ogc.geometry.simple.LinearRing;
import de.hsbo.fbv.ogc.geometry.simple.MultiLineString;
import de.hsbo.fbv.ogc.geometry.simple.MultiPoint;
import de.hsbo.fbv.ogc.geometry.simple.MultiPolygon;
import de.hsbo.fbv.ogc.geometry.simple.Point;
import de.hsbo.fbv.ogc.geometry.simple.Polygon;

/**
 * Implementierung des Simple Feature Modells auf Basis der Klassenbibliothek
 * Java Topology Suite (JTS) 1.10 der Fa. VividSolutions.<br>
 * (C) siehe Lizenzen von Geotools <br>
 * Factory-Klasse fr verschiedene geometrische Formen. Zugleich steuert die 
 * Klasse zentrale Parameter.
 * 
 * @author A.Zimmermann, HS Bochum, 2011
 * @version 0.5
 */
public class GeoModel {

	/** Konstante DEFAULT entspricht der Standardeinstellung */
	public static final int DEFAULT = -1;
	/** Konstante X: Index fr x-Komponente */	
	public static final int X = 0;
	/** Konstante Y: Index fr y-Komponente */	
	public static final int Y = 1;
	/** Konstante Z: Index fr z-Komponente */	
	public static final int Z = 2;
	/** Konstante M: Index fr m-Komponente */	
	public static final int M = 3;
	/** Verweis auf die Factory-Klasse der Java Topology Suite (JTS) */
	public static final GeometryFactory FACTORY = new GeometryFactory();

	/** Klassenattribut DEFAULT */
	static int sidx = DEFAULT;
	/** Klassenattribut EPS: Toleranzgrenze fr geometrische Auflsung */
	static double EPS = 0.001;
	/** Klassenattribut sortIndex: Komponente zum Sortieren der Objekte */
	static int sortIndex = GeoModel.DEFAULT;

	/**
	 * Getter-Methode fr EPS
	 * @return
	 */
	public static double getEPS() {
		return EPS;
	}

	/**
	 * Setter-Methode fr EPS
	 * @param eps
	 */
	public static void setEPS(double eps) {
		EPS = eps;
	}

	/**
	 * Getter-Methode fr sortIndex
	 * @return
	 */
	public static int getSortIndex() {
		return sidx;
	}

	/**
	 * Setter-Methode fr sortIndex
	 * @param sidx
	 */
	public static void setSortIndex(int sidx) {
		if (sidx >= DEFAULT && sidx <= M)
			GeoModel.sidx = sidx;
	}

	/**
	 * Factory-Methode fr GPoint bzw. Point
	 * @param x : x-Wert
	 * @param y : y-Wert
	 * @return GPoint
	 */
	public static GPoint createPoint(double x, double y) {
		try {
			return new GPoint(x, y, Double.NaN, Double.NaN);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GPoint bzw. Point
	 * @param x : x-Wert
	 * @param y : y-Wert
	 * @param m : m-Wert
	 * @return GPoint
	 */
	public static GPoint createPointM(double x, double y, double m) {
		try {
			return new GPoint(x, y, Double.NaN, m);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GPoint bzw. Point
	 * @param x : x-Wert
	 * @param y : y-Wert
	 * @param z : z-Wert
	 * @return GPoint
	 */
	public static GPoint createPoint(double x, double y, double z) {
		try {
			return new GPoint(x, y, z, Double.NaN);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GPoint bzw. Point
	 * @param x : x-Wert
	 * @param y : y-Wert
	 * @param z : z-Wert
	 * @param m : m-Wert
	 * @return GPoint
	 */
	public static GPoint createPointM(double x, double y, double z, double m) {
		try {
			return new GPoint(x, y, z, m);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GPoint bzw. Point
	 * @param c : CoordinateM-Objekt
	 * @return GPoint
	 */
	public static GPoint createPoint(CoordinateM c) {
		try {
			return new GPoint(c);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GPoint bzw. Point
	 * @param p : Objekt vom Interface-Typ Point
	 * @return GPoint
	 */
	public static GPoint createPoint(Point p) {
		return createPointM(p.x(), p.y(), p.z(), p.m());
	}

	/**
	 * Factory-Methode fr GPoint bzw. Point
	 * @param p : Point-Objekt der Java Topology Suite
	 * @return GPoint
	 */
	public static GPoint createPoint(com.vividsolutions.jts.geom.Point p) {
		try {
			CoordinateM mc = new CoordinateM(p.getCoordinate());
			return new GPoint(mc.x, mc.y, mc.z, mc.m);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GLineString bzw. LineString
	 * @param p : Feld vom Typ GPoint
	 * @return GLineString
	 */
	public static GLineString createLineString(GPoint[] p) {
		try {
			return new GLineString(p);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GLineString bzw. LineString
	 * @param c : Feld vom Typ CoordinateM
	 * @return GLineString
	 */
	public static GLineString createLineString(CoordinateM[] c) {
		try {
			return new GLineString(c);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GLineString bzw. LineString
	 * @param l : Objekt vom Interface-Typ LineString
	 * @return GLineString
	 */
	public static GLineString createLineString(LineString l) {
		try {
			GPoint[] points = new GPoint[l.numPoints()];
			for (int i = 0; i < points.length; i++)
				points[i] = createPoint(l.pointN(i));
			return new GLineString(points);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GLineString bzw. LineString
	 * @param l : LineString-Objekt der Java Topology Suite
	 * @return GLineString
	 */
	public static GLineString createLineString(
			com.vividsolutions.jts.geom.LineString l) {
		try {
			Coordinate[] coords = l.getCoordinates();
			CoordinateM[] mcoords = new CoordinateM[l.getCoordinates().length];
			for (int i = 0; i < mcoords.length; i++)
				mcoords[i] = new CoordinateM(coords[i]);
			return new GLineString(mcoords);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GLine bzw. Line
	 * @param ps : Startpunkt (Typ GPoint)
	 * @param pe : Endpunkt (Typ GPoint)
	 * @return GLine
	 */
	public static GLine createLine(GPoint ps, GPoint pe) {
		try {
			return new GLine(ps, pe);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GLine bzw. Line
	 * @param cs : Startkoordinate (Typ CoordinateM)
	 * @param ce : Endkoordinate (Typ CoordinateM)
	 * @return GLine
	 */
	public static GLine createLine(CoordinateM cs, CoordinateM ce) {
		try {
			return new GLine(cs, ce);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GLine bzw. Line
	 * @param l : Objekt vom Interface-Typ Line
	 * @return GLine
	 */
	public static GLine createLine(Line l) {
		try {
			return new GLine(createPoint(l.startPoint()),
					createPoint(l.endPoint()));
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GLine bzw. Line
	 * @param l : LineSegment-Objekt der Java Topology Suite
	 * @return GLine
	 */
	public static GLine createLine(com.vividsolutions.jts.geom.LineSegment l) {
		try {
			return new GLine(new CoordinateM(l.p0), new CoordinateM(l.p1));
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GLinearRing bzw. LinearRing
	 * @param p : Feld vom Typ GPoint
	 * @return GLinearRing
	 */
	public static GLinearRing createLinearRing(GPoint[] p) {
		try {
			return new GLinearRing(p);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GLinearRing bzw. LinearRing
	 * @param c : Feld vom Typ CoordinateM
	 * @return GLinearRing
	 */
	public static GLinearRing createLinearRing(CoordinateM[] c) {
		try {
			return new GLinearRing(c);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GLinearRing bzw. LinearRing
	 * @param r : Objekt vom Interface-Typ LinearRing
	 * @return GLinearRing
	 */
	public static GLinearRing createLinearRing(LinearRing r) {
		try {
			return createLineString(r).toLinearRing();
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GLinearRing bzw. LinearRing
	 * @param r : LinearRing-Objekt der Java Topology Suite
	 * @return GLinearRing
	 */
	public static GLinearRing createLinearRing(
			com.vividsolutions.jts.geom.LinearRing r) {
		try {
			return createLineString(r).toLinearRing();
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GLinearRing bzw. LinearRing
	 * @param l : Objekt vom Interface-Typ LineString
	 * @return GLinearRing
	 */
	public static GLinearRing createLinearRing(LineString l) {
		try {
			return createLineString(l).toLinearRing();
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}
	
	/**
	 * Factory-Methode fr GLinearRing bzw. LinearRing
	 * @param l : LinearRing-Objekt der Java Topology Suite
	 * @return GLinearRing
	 */
	public static GLinearRing createLinearRing(
			com.vividsolutions.jts.geom.LineString l) {
		try {
			return createLineString(l).toLinearRing();
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GPolygon bzw. Polygon
	 * @param shell : uere Begrenzungslinie (Typ GLinearRing)
	 * @param holes : Feld der innere Begrenzungslinien (Typ GLinearRing)
	 * @return GPolygon
	 */
	public static GPolygon createPolygon(GLinearRing shell, GLinearRing[] holes) {
		try {
			return new GPolygon(shell, holes);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GPolygon bzw. Polygon
	 * @param p : Objekt vom Interface-Typ Polygon
	 * @return GPolygon
	 */
	public static GPolygon createPolygon(Polygon p) {
		try {
			GLinearRing shell = GeoModel.createLinearRing(p.exteriorRing());
			GLinearRing[] holes = new GLinearRing[p.numInteriorRing()];
			for (int i = 0; i < holes.length; i++) {
				holes[i] = GeoModel.createLinearRing(p.interiorRingN(i));
			}
			return new GPolygon(shell, holes);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GPolygon bzw. Polygon
	 * @param p : Polygon-Objekt der Java Topology Suite
	 * @return GPolygon
	 */
	public static GPolygon createPolygon(com.vividsolutions.jts.geom.Polygon p) {
		try {
			GLinearRing shell = GeoModel.createLinearRing(p.getExteriorRing());
			GLinearRing[] holes = new GLinearRing[p.getNumInteriorRing()];
			for (int i = 0; i < holes.length; i++) {
				holes[i] = GeoModel.createLinearRing(p.getInteriorRingN(i));
			}
			return new GPolygon(shell, holes);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GMultiPoint bzw. MultiPoint
	 * @param mpt : Objekt vom Interface-Typ MultiPoint
	 * @return GMultiPoint
	 */
	public static GMultiPoint createMultiPoint(MultiPoint mpt) {
		try {
			GPoint[] geoms = new GPoint[mpt.numGeometries()];
			for (int i = 0; i < geoms.length; i++)
				geoms[i] = createPoint((Point) mpt.geometryN(i));
			return new GMultiPoint(geoms);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GMultiPoint bzw. MultiPoint
	 * @param mpt : Multi-Point-Objekt der Java Topology Suite
	 * @return GMultiPoint
	 */
	public static GMultiPoint createMultiPoint(
			com.vividsolutions.jts.geom.MultiPoint mpt) {
		try {
			GPoint[] geoms = new GPoint[mpt.getNumGeometries()];
			for (int i = 0; i < geoms.length; i++)
				geoms[i] = createPoint((com.vividsolutions.jts.geom.Point) mpt
						.getGeometryN(i));
			return new GMultiPoint(geoms);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GMultiLineString bzw. MultiLineString
	 * @param mls : Objekt vom Interface-Typ MultiLineString
	 * @return GMultiLineString
	 */
	public static GMultiLineString createMultiLineString(MultiLineString mls) {
		try {
			GLineString[] geoms = new GLineString[mls.numGeometries()];
			for (int i = 0; i < geoms.length; i++)
				geoms[i] = createLineString((com.vividsolutions.jts.geom.LineString) mls
						.geometryN(i));
			return new GMultiLineString(geoms);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GMultiLineString bzw. MultiLineString
	 * @param mls : MultiLineString-Objekt der Java Topology Suite
	 * @return GMultiLineString
	 */
	public static GMultiLineString createMultiLineString(
			com.vividsolutions.jts.geom.MultiLineString mls) {
		try {
			GLineString[] geoms = new GLineString[mls.getNumGeometries()];
			for (int i = 0; i < geoms.length; i++)
				geoms[i] = createLineString((LineString) mls.getGeometryN(i));
			return new GMultiLineString(geoms);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GMultiPolygon bzw. MultiPolygon
	 * @param mpl : Objekt vom Interface-Typ MultiPolygon
	 * @return GMultiPolygon
	 */
	public static GMultiPolygon createMultiPolygon(MultiPolygon mpl) {
		try {
			GPolygon[] geoms = new GPolygon[mpl.numGeometries()];
			for (int i = 0; i < geoms.length; i++)
				geoms[i] = createPolygon((Polygon) mpl.geometryN(i));
			return new GMultiPolygon(geoms);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GMultiPolygon bzw. MultiPolygon
	 * @param mpl : MultiPolygon-Objekt der Java Topology Suite
	 * @return GMultiPolygon
	 */
	public static GMultiPolygon createMultiPolygon(
			com.vividsolutions.jts.geom.MultiPolygon mpl) {
		try {
			GPolygon[] geoms = new GPolygon[mpl.getNumGeometries()];
			for (int i = 0; i < geoms.length; i++)
				geoms[i] = GeoModel
						.createPolygon((com.vividsolutions.jts.geom.Polygon) mpl
								.getGeometryN(i));
			return new GMultiPolygon(geoms);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GGeometry bzw. Geometry
	 * @param geom : Objekt vom Interface-Typ Geometry
	 * @return GGeometry
	 * @throws Exception Geometry-Datentyp ist nicht implementiert
	 */
	public static GGeometry createGeometry(Geometry geom) throws Exception {
		if (geom instanceof Point)
			return createPoint((Point) geom);
		else if (geom instanceof LineString)
			return createLineString((LineString) geom);
		else if (geom instanceof Polygon)
			return createPolygon((Polygon) geom);
		else if (geom instanceof Line)
			return createLine((Line) geom);
		else if (geom instanceof LinearRing)
			return createLinearRing((LinearRing) geom);
		else if (geom instanceof MultiPoint)
			return createMultiPoint((MultiPoint) geom);
		else if (geom instanceof MultiLineString)
			return createMultiLineString((MultiLineString) geom);
		else if (geom instanceof MultiPolygon)
			return createMultiPolygon((MultiPolygon) geom);
		else if (geom instanceof GeometryCollection)
			return createGeometryCollection((GeometryCollection) geom);
		else {
			throw new Exception("Geometry type not implemented!");
		}
	}

	/**
	 * Factory-Methode fr GGeometry bzw.Geometry-Objekten 
	 * @param geom : Geometrie-Objekt der Java Topology Suite (JTS)
	 * @return GGeometry
	 * @throws Exception Geometry-Datentyp ist nicht implementiert
	 */
	public static GGeometry createGeometry(
			com.vividsolutions.jts.geom.Geometry geom) throws Exception {
		if (geom instanceof com.vividsolutions.jts.geom.Point)
			return createPoint((com.vividsolutions.jts.geom.Point) geom);
		else if (geom instanceof com.vividsolutions.jts.geom.LineString)
			return createLineString((com.vividsolutions.jts.geom.LineString) geom);
		else if (geom instanceof com.vividsolutions.jts.geom.Polygon)
			return createPolygon((com.vividsolutions.jts.geom.Polygon) geom);
		else if (geom instanceof com.vividsolutions.jts.geom.LinearRing)
			return createLinearRing((com.vividsolutions.jts.geom.LinearRing) geom);
		else if (geom instanceof com.vividsolutions.jts.geom.MultiPoint)
			return createMultiPoint((com.vividsolutions.jts.geom.MultiPoint) geom);
		else if (geom instanceof com.vividsolutions.jts.geom.MultiLineString)
			return createMultiLineString((com.vividsolutions.jts.geom.MultiLineString) geom);
		else if (geom instanceof com.vividsolutions.jts.geom.MultiPolygon)
			return createMultiPolygon((com.vividsolutions.jts.geom.MultiPolygon) geom);
		else if (geom instanceof com.vividsolutions.jts.geom.GeometryCollection)
			return createGeometryCollection((com.vividsolutions.jts.geom.GeometryCollection) geom);
		else {
			throw new Exception("Geometry type not implemented!");
		}
	}

	/**
	 * Factory-Methode fr GGeometryCollection bzw. GeometryCollection
	 * @param gc : Objekt vom Interface-Typ GeometryCollection
	 * @return GGeometryCollection
	 */
	public static GGeometryCollection createGeometryCollection(
			GeometryCollection gc) {
		try {
			GGeometry[] geoms = new GGeometry[gc.numGeometries()];
			for (int i = 0; i < geoms.length; i++)
				geoms[i] = createGeometry(gc.geometryN(i));
			return new GGeometryCollection(geoms);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GGeometryCollection bzw. GeometryCollection
	 * @param gc : GeometrieCollection-Objekt der Java Topology Suite
	 * @return GGeometryCollection
	 */	
	public static GGeometryCollection createGeometryCollection(
			com.vividsolutions.jts.geom.GeometryCollection gc) {
		try {
			GGeometry[] geoms = new GGeometry[gc.getNumGeometries()];
			for (int i = 0; i < geoms.length; i++)
				geoms[i] = createGeometry(gc.getGeometryN(i));
			return new GGeometryCollection(geoms);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Factory-Methode fr GTriangle bzw. Triangle
	 * @param a : Punkt A
	 * @param b : Punkt B
	 * @param c : Punkt C
	 * @return GTriangle
	 */
	public static GTriangle createTriangle(GPoint a, GPoint b, GPoint c) {
		try {
			return new GTriangle(a.getCoordinate(), b.getCoordinate(),
					c.getCoordinate());
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return null;
		}
	}

} // Ende Klasse
