package de.hsbo.fbv.bmg.geometry.simple.collections;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;

import de.hsbo.fbv.bmg.geometry.simple.CoordinateM;
import de.hsbo.fbv.bmg.geometry.simple.GPoint;
import de.hsbo.fbv.bmg.geometry.simple.GeoModel;



/**
 * Hilfsklasse zur Behandlung von Punkten als statische Felder.<br>
 * Die Klasse lsst die Speicherung von Punkt-Duplikaten zu. Die 
 * Punktfelder knnen zwar ausgewertet, jedoch nicht verndert werden.
 * Insbesondere knnen keine Punkte ergnzt oder entfernt werden.
 * Die Index-Reihenfolge der Elemente ist abhngig von den vorausgegangenen 
 * Operationen.
 * @author zimmermann
 *
 */
public class GPointArray {
	
	// Parameter
	public static final int minx = 0, W = 0, SW = 0;
	public static final int miny = 1, S = 1, SE = 1;
	public static final int maxx = 2, E = 2, NW = 2;
	public static final int maxy = 3, N = 3, NE = 3;

	// Attribute
	GPoint[] array = null;

	/**
	 * Konstruktor mit Angabe eines Feldes. Speichert eine Kopie des Feldes.
	 * @param array
	 */
	public GPointArray(GPoint[] array) {
		super();
		this.array = Arrays.copyOf(array, array.length);
	}
	
	/**
	 * Kopier-Konstruktor
	 * @param array
	 */
	public GPointArray(GPointArray array) {
		this.array = toArray();
	}
	
	/**
	 * Ruft das Element mit dem Index i ab!
	 * @param i
	 * @return
	 */
	public GPoint get(int i) {
		return array[i];
	}
	
	/**
	 * Berechnet den Lageschwerpunkt des Feldes.
	 * @return
	 */
	public GPoint getCenter() {
		double x=0, y=0;
		int n = array.length;
		for(int i=0; i<n; i++) {
			x = x + array[i].x();
			y = y + array[i].y();
		}
		return GeoModel.createPoint(x/n, y/n);
	}
	
	/**
	 * Gibt die Feldlnge an.
	 * @return
	 */
	public int size() {
		return array.length;
	}	
	
	/**
	 * Gibt eine Kopie des Feldes aus.
	 * @return
	 */
	public GPoint[] toArray() {
		return Arrays.copyOf(array, array.length);
	}
	
	/**
	 * Sortiert nach der Komponente (Indexe der Komponenten in GeoModel).
	 * @param sidx
	 */
	public void sort(int sidx) {
		if(sidx>=GeoModel.DEFAULT && sidx<=GeoModel.M) {
			Arrays.sort(array, new GPointComp(sidx));
		}
	}
	
	/**
	 * Teilt das Punktfeld am Wert x in zwei neue Punktfelder auf.
	 * Alle Punkte mit p.x()<=x sind im ersten Punktfeld, die greren im zweiten Feld.
	 * @param x
	 * @return
	 */
	public GPointArray[] splitX(double x) {
		this.sort(GeoModel.X);
		int ix = this.searchNext(GeoModel.createPoint(x, Double.NaN), GeoModel.X);
		return split(ix);
	}
	
	/**
	 * Teilt das Punktfeld am Wert y in zwei neue Punktfelder auf.
	 * Alle Punkte mit p.y()<=y sind im ersten Punktfeld, die greren im zweiten Feld.
	 * @param y
	 * @return
	 */
	public GPointArray[] splitY(double y) {
		this.sort(GeoModel.Y);
		int ix = this.searchNext(GeoModel.createPoint(Double.NaN, y), GeoModel.Y);
		return split(ix);
	}
	
	/**
	 * Teilt das Punktfeld an der Koordinate (x,y) in vier neue Punktfelder auf.
	 * @param x
	 * @param y
	 * @return
	 */
	public GPointArray[] splitXY(double x, double y) {
		GPointArray[] res = new GPointArray[4];
		// Split in x-Richtung
		GPointArray[] s = this.splitX(x);
		if(s[0]==null) {
			res[SW] = null;
			res[NW] = null;
		}
		else  {
			GPointArray[] sw = s[0].splitY(y);
			res[SW] = sw[0];
			res[NW] = sw[1];			
		}
		if(s[1]==null) {
			res[SE] = null;
			res[NE] = null;
		}
		else  {
			GPointArray[] se = s[1].splitY(y);
			res[SE] = se[0];
			res[NE] = se[1];			
		}
		return res;
	}
	
	/**
	 * Gibt alle Punkte zurck, die im angegebenen Rechteck liegen.
	 * @param x
	 * @param y
	 * @param dx
	 * @param dy
	 * @return
	 */
	public GPointArray selectRectangle(double x, double y, double dx, double dy) {
		GPointArray pa = splitXY(x,y)[NE];
		return pa.splitXY(x+dx, y+dy)[SW];
	}
	
	/**
	 * Gibt alle Punkte zurck, die im angegebenen Rechteck liegen.
	 * @param a
	 * @param b
	 * @return
	 */
	public GPointArray selectRectangle(GPoint a, GPoint b) {
		double xmin = Math.min(a.x(), b.x());
		double ymin = Math.min(a.y(), b.y());
		double xmax = Math.max(a.x(), b.x());
		double ymax = Math.max(a.y(), b.y());
		ArrayList<GPoint> al = new ArrayList<GPoint>();
		for(int i=0; i<this.array.length; i++) {
			GPoint p = this.array[i];
			if(p.x()>xmin && p.x()<xmax && p.y()>ymin && p.y()<ymax)
				al.add(p);
		}
		
		if (!al.isEmpty()) {
			GPoint[] pa = new GPoint[al.size()];
			return new GPointArray(al.toArray(pa));
		}
		else 
			return null;
		/*
		GPointArray pa = splitXY(xmin,ymin)[NE];
		return pa.splitXY(xmax, ymax)[SW];
		*/
	}
	
	/**
	 * Gibt die extremen Punkte des Feldes an, durch die die BoundingBox bestimmt ist.
	 * Die Punkte sind in der Reihenfolge W,S,E,N angegeben.
	 * @return
	 */
	public GPoint[] getExtremePointsXY() {
		GPoint[] result = new GPoint[4];
		this.sort(GeoModel.X);
		result[minx] = array[0];
		result[maxx] = array[array.length-1];
		this.sort(GeoModel.Y);
		result[miny] = array[0];
		result[maxy] = array[array.length-1];
		return result;
	}
	
	/**
	 * Gibt die Koordinatenwerte der BoundingBox in der Reihenfolge W,S,E,N an.
	 * @return
	 */
	public double[] getBoundsXY() {
		double[] result = new double[4];
		this.sort(GeoModel.X);
		result[minx] = array[0].x();
		result[maxx] = array[array.length-1].x();
		this.sort(GeoModel.Y);
		result[miny] = array[0].y();
		result[maxy] = array[array.length-1].y();
		return result;
	}
	
	/** 
	 * Bestimmt den zur Koordinate (x,y) nchstgelegenen Punkt.
	 * @param x
	 * @param y
	 * @return
	 */
	public GPoint nearestPoint(double x, double y) {
		GPoint p = GeoModel.createPoint(x, y);
		this.sort(GeoModel.X);
		int inp = this.searchNext(p, GeoModel.X);
		GPoint np = this.get(inp);
		GPoint res = np;
		if(p.equals(np))
			return res;
		else {
			double dmin = p.distance(np);
			int i = inp++;
			// Suche fr alle Punkte mit Index i>inp
			for(;i<array.length && Math.abs(x-get(i).x())<dmin; i++) {
				if(p.distance(get(i))<dmin) {
					res = get(i);
					dmin = p.distance(res);
				}
			}
			i = inp--;
			// Suche fr alle Punkte mit Index i<inp
			for(;i<=0 && Math.abs(x-get(i).x())<dmin; i--) {
				if(p.distance(get(i))<dmin) {
					res = get(i);
					dmin = p.distance(res);
				}
			}
			return res;
		}
	}
	
	/**
	 * Bestimmt den zum Punkt p nchstgelegenen Punkt.
	 * @param p
	 * @return
	 */
	public GPoint nearestPoint(GPoint p) {
		return nearestPoint(p.x(), p.y());
	}
	
	/**
	 * Gibt die Eckpunkte fr ein Rechteckeck an, um das Zentrum der Punktemenge liegt.<br>
	 * Mit dem Faktor f  kann das Rechteck um das Zentrum vergrert werden.
	 * mit dem Faktor f=1.0 wird die BoundingBox (Envelope) erzeugt.
	 * @param f
	 * @return
	 */
	public GPoint[] getABox(double f) {
		double[] bbox = getBoundsXY();
		double cx = (bbox[E]+bbox[W])/2;
		double sx = (bbox[E]-bbox[W])/2;
		double cy = (bbox[N]+bbox[S])/2;
		double sy = (bbox[N]-bbox[S])/2;
		GPoint[] res = new GPoint[4];
		res[SW] = GeoModel.createPoint(cx-f*sx, cy-f*sy);
		res[SE] = GeoModel.createPoint(cx+f*sx, cy-f*sy);
		res[NW] = GeoModel.createPoint(cx-f*sx, cy+f*sy);
		res[NE] = GeoModel.createPoint(cx+f*sx, cy+f*sy);
		return res;
	}
	
	/**
	 * Ausgabe
	 */
	public String toString() {
		return Arrays.toString(array);
	}
	
	private int searchNext(GPoint point, int sidx) {
		int res = Arrays.binarySearch(array, point, new GPointComp(sidx));
		if(res<0)
			return Math.abs(res)-1;
		else {
			GPointComp compare = new GPointComp(sidx);
			while(res<array.length-1 && compare.compare(array[res], array[res+1])==0)
				res++;
			res++;
		}	
		return res;
	}
		
	/**
	 * Interne Methode
	 * @param i
	 * @return
	 */
	private GPointArray[] split(int i) {
		if(i>0 && i<array.length) {
			return new GPointArray[] {
					new GPointArray(Arrays.copyOfRange(array, 0, i)),
					new GPointArray(Arrays.copyOfRange(array, i, array.length))};			
		}
		else if(i==0) {
			return new GPointArray[] {
					null, new GPointArray(Arrays.copyOfRange(array, i, array.length))};
		}
		else if(i==array.length) {
			return new GPointArray[] {
					new GPointArray(Arrays.copyOfRange(array, 0, array.length)), null};
		}
		else 
			return null;
	}
	
}	// Ende Klasse GPointArray

/**
 * Hilfsklasse: Komparator fr GPoint
 * @author zimmermann
 *
 */
class GPointComp implements Comparator<GPoint> {
	
	int sidx = GeoModel.DEFAULT;
	
	public GPointComp() {}
	
	public GPointComp(int sidx){
		this.sidx = sidx;
	}

	@Override
	public int compare(GPoint o1, GPoint o2) {
		return o1.compareTo(o2, sidx);
	}
	
}	// Ende KLasse GPointComp
