package de.hsbo.fbv.ogc.geometry.simple;

/**
 * (Abstrakte) Urklasse (in Java als Interface) von Simple Feature Access - 
 * Common Architecture (SFA-CA). 
 * Die instanziierbaren Kindklassen sind beschrnkt auf maximal 2,5D-Objekte 
 * im 2- bis 4-dimensionalen Objektraum.
 * @see Dokument OGC 06-103r3
 * @author (OGC)
 * @version 1.2
 */
public interface Geometry {
	
	/**
	 * Assoziation (Beziehung) zum rumlichen Referenzsystem (Kardinalitt: 1)
	 */
	// ReferenceSystem spatialRS;
	
	/**
	 * Assoziation (Beziehung) zum Referenzsystem des Messwertes (Kardinalitt: 0..1)
	 */
	// ReferenceSystem measureRS;

	/**
	 * Gruppe A (Grundoperationen):<br> 
	 * Gibt die geometrische (topologische) Dimension des Objektes an. 
	 * @return :int
	 */
	public int dimension();

	/**
	 * Gruppe A (Grundoperationen):<br> 
	 * Gibt die Dimension des verwendeten Koordinatensystems an. 
	 * @return :int
	 */
	public int coordinateDimension();
	
	/**
	 * Gruppe A (Grundoperationen):<br> 
	 * Gibt die rmliche Dimension des Geometrieobjektes an. 
	 * @return :int
	 */
	public int spatialDimension();
	
	/**
	 * Gruppe A (Grundoperationen):<br> 
	 * Gibt des Namen der (instanziierbaren) Kindklasse an.
	 * @return :String
	 */
	public String geometryType();
	
	/**
	 * Gruppe A (Grundoperationen):<br> 
	 * Gibt die ID des rumlichen Referenzsystems an.
	 * @return :int
	 */
	public int srid();
	
	/**
	 * Gruppe A (Grundoperationen):<br> 
	 * Gibt das umschreibende Rechteck (bounding box) des Objektes an.
	 * Das Ergebnis ist ein Polygon, das die Eckpunkte des Rechtecks enthlt.
	 * @return :Geometry
	 */
	public Geometry envelope();
	
	/**
	 * Gruppe A (Grundoperationen):<br> 
	 * Exportiert das Objekt als Well-known Text (WKT).
	 * @return :String
	 */
	public String asText();

	/**
	 * Gruppe A (Grundoperationen):<br> 
	 * Exportiert das Objekt als Well-known Binary (WKB).
	 * @return :Binary
	 */
	public Binary asBinary();
	
	/**
	 * Gruppe A (Grundoperationen):<br> 
	 * Untersucht, ob das Objekt eine leere Menge enthlt.
	 * Das Ergebnis ist true, wenn im Objekt keine Punkte vorhanden sind, sonst false.
	 * @return :boolean
	 */
	public boolean isEmpty();
	
	/**
	 * Gruppe A (Grundoperationen):<br> 
	 * Untersucht, ob das Objekt keine geomertrische Anomalien besitzt.
	 * Das Ergebnis ist true, wenn keine berschneidungen und tangentialen Berhrungen seiner 
	 * Segmente vorhanden sind, sonst false.
	 * @return :boolean
	 */
	public boolean isSimple();
	
	/**
	 * Gruppe A (Grundoperationen):<br> 
	 * Untersucht, ob das Objekt z-Koordinatenwerte enthlt (-> true). 
	 * @return :boolean
	 */
	public boolean is3D();
	
	/**
	 * Gruppe A (Grundoperationen):<br> 
	 * Untersucht, ob das Objekt m-Werte enthlt (-> true).
	 * @return :boolean
	 */
	public boolean isMeasured();
	
	/**
	 * Gruppe A (Grundoperationen):<br> 
	 * Gibt den Rand, d.h. die begrenzende Geometrie (Boundary) des Objektes zurck. 
	 * @return :Geometry
	 */
	public Geometry boundary();
	
	/**
	 * Gruppe B (Beziehungen zu anderen Geometrie-Objekten):<br> 
	 * Untersucht, ob die Geometrien rumlich identisch sind (-> true).
	 * @param anotherGeometry
	 * @return :boolean
	 */
	public boolean equals(Geometry anotherGeometry);
	
	/**
	 * Gruppe B (Beziehungen zu anderen Geometrie-Objekten):<br> 
	 * Untersucht, ob die beiden Geometrien rumlich voneinander entfernt (disjoint) 
	 * liegen (-> true).
	 * @param anotherGeometry
	 * @return :boolean
	 */
	public boolean disjoint(Geometry anotherGeometry);
	
	/**
	 * Gruppe B (Beziehungen zu anderen Geometrie-Objekten):<br> 
	 * Untersucht, ob sich die beiden Geometrien berlagern (->true).
	 * @param anotherGeometry
	 * @return :boolean
	 */
	public boolean intersects(Geometry anotherGeometry);
	
	/**
	 * Gruppe B (Beziehungen zu anderen Geometrie-Objekten):<br> 
	 * Untersucht, ob sich die beiden Geometrien berhren (->true).	
	 * @param anotherGeometry
	 * @return :boolean
	 */
	public boolean touches(Geometry anotherGeometry);
	
	/**
	 * Gruppe B (Beziehungen zu anderen Geometrie-Objekten):<br> 
	 * Untersucht, ob sich die beiden Geometrien schneiden (->true).
	 * @param anotherGeometry
	 * @return :boolean
	 */
	public boolean crosses(Geometry anotherGeometry);
	
	/**
	 * Gruppe B (Beziehungen zu anderen Geometrie-Objekten):<br> 
	 * Untersucht, ob die Geometrie 'this' rumlich innerhalb von 'anotherGeometry' liegt (->true)
	 * @param anotherGeometry
	 * @return :boolean
	 */
	public boolean within(Geometry anotherGeometry);
	
	/**
	 * Gruppe B (Beziehungen zu anderen Geometrie-Objekten):<br> 
	 * Untersucht, ob die Geometrie 'this' die Geometrie 'anotherGeometry' rumlich enthlt (->true)
	 * @param anotherGeometry
	 * @return :boolean
	 */
	public boolean contains(Geometry anotherGeometry);

	/**
	 * Gruppe B (Beziehungen zu anderen Geometrie-Objekten):<br> 
	 * Untersucht, ob die Geometrien sich rumlich berlappen (->true)
	 * @param anotherGeometry
	 * @return :boolean
	 */	
	public boolean overlaps(Geometry anotherGeometry);
	
	/**
	 * Gruppe B (Beziehungen zu anderen Geometrie-Objekten):<br> 
	 * Untersucht, ob die 'IntersectionPatternMatrix' DE-9IM der beiden Geometrien 
	 * eine rumliche Beziehung aufweist (->true).
	 * @param anotherGeometry
	 * @param intersectionPatternMatrix
	 * @return :boolean
	 */
	public boolean relate(Geometry anotherGeometry, String intersectionPatternMatrix);
	
	/**
	 * Gruppe B (Beziehungen zu anderen Geometrie-Objekten):<br> 
	 * Gibt die Teile einer Geometrie zurck, deren M-Komponenten mit mValue bereinstimmen.
	 * @param mValue
	 * @return :Geometry (MultiPoint oder MultiLineString)
	 */
	public Geometry locateAlong(double mValue);
	
	/**
	 * Gruppe B (Beziehungen zu anderen Geometrie-Objekten):<br> 
	 * Gibt die Teile einer Geometrie zurck, deren M-Komponenten zwischen mStart 
	 * und mEnd liegen.
	 * @param mStart
	 * @param mEnd
	 * @return :Geometry (MultiPoint oder MultiLineString)
	 */
	public Geometry locateBetween(double mStart, double mEnd);
	
	/**
	 * Gruppe C (Raumbezogene Analyse):<br> 
	 * Berechnet den krzesten Abstand zwischen den angegebenen Geometrien.
	 * @param anotherGeometry
	 * @return :double
	 */
	public double distance(Geometry anotherGeometry);
	
	/**
	 * Gruppe C (Raumbezogene Analyse):<br> 
	 * Berechnet eine Geometrie, die alle Punkte im Abstand distance enthlt.
	 * @param distance
	 * @return :Geometry (Polygon)
	 */
	public Geometry buffer(double distance);
	
	/**
	 * Gruppe C (Raumbezogene Analyse):<br> 
	 * Berechnet die konvexe Hlle der angegebenen Geometrie.
	 * @return :Geometry (Polygon)
	 */
	public Geometry convexHull();
	
	/**
	 * Gruppe C (Raumbezogene Analyse):<br> 
	 * Berechnet die Schnittmenge der angegebenen Geometrien.
	 * @param anotherGeometry
	 * @return :Geometry
	 */
	public Geometry intersection(Geometry anotherGeometry);
	
	/**
	 * Gruppe C (Raumbezogene Analyse):<br> 
	 * Berechnet die Vereinigungsmenge der angegebenen Geometrien.
	 * @param anotherGeometry
	 * @return :Geometry
	 */
	public Geometry union(Geometry anotherGeometry);
	
	/**
	 * Gruppe C (Raumbezogene Analyse):<br> 
	 * Berechnet die Restmenge der angegebenen Geometrie, von der 'anotherGeometry' 
	 * abgezogen wird.
	 * @param anotherGeometry
	 * @return :Geometry
	 */
	public Geometry difference(Geometry anotherGeometry);
	
	/**
	 * Gruppe C (Raumbezogene Analyse):<br> 
	 * Berechnet die Restmenge, die durch den gegenseitigen Abzug beider Geometrien entsteht.
	 * @param anotherGeometry
	 * @return :Geometry
	 */
	public Geometry symDifference(Geometry anotherGeometry);
	
}
